from .audio_codec import AudioCodec
from .prompt_processor import PromptProcessor
from .model import HFModel, GGUFModel, GenerationConfig
import torch
import sounddevice as sd
from .alignment import CTCForcedAlignment
import torchaudio
import tempfile
from dataclasses import dataclass

@dataclass
class ModelOutput:
    audio: torch.Tensor
    sr: int

    def save(self, path: str):
        torchaudio.save(path, self.audio.cpu(), sample_rate=self.sr, encoding='PCM_S', bits_per_sample=16)

    def play(self):
        sd.play(self.audio[0].cpu().numpy(), self.sr)
        sd.wait()

class InterfaceHF:
    def __init__(
        self,
        model_path: str,
        device: str = None,
        dtype: torch.dtype = None,
        additional_model_config: dict = {}
    ) -> None:
        self.device = torch.device(
            device if device is not None 
            else "cuda" if torch.cuda.is_available() 
            else "cpu"
        )
        self.audio_codec = AudioCodec(self.device)
        self.prompt_processor = PromptProcessor(model_path)
        self.model = HFModel(model_path, self.device, dtype, additional_model_config)

    def prepare_prompt(self, text: str, speaker: dict = None):
        prompt = self.prompt_processor.get_completion_prompt(text, speaker)
        return self.prompt_processor.tokenizer.encode(
            prompt, add_special_tokens=False, return_tensors="pt").to(self.model.device)

    def get_audio(self, tokens):
        output = self.prompt_processor.extract_audio_from_tokens(tokens)
        return self.audio_codec.decode(
            torch.tensor([[output]], dtype=torch.int64).to(self.audio_codec.device)
        )

    def create_speaker(self, audio_path: str, transcript: str):
        ctc = CTCForcedAlignment()
        words = ctc.align(audio_path, transcript)
        ctc.free()

        full_audio = torch.cat([i["audio"] for i in words], dim=1)

        with tempfile.NamedTemporaryFile(suffix=".wav") as temp_file:
            torchaudio.save(temp_file.name, full_audio, ctc.sample_rate)
            full_codes = self.audio_codec.encode(self.audio_codec.load_audio(temp_file.name)).tolist()

        data = []
        start = 0

        for i in words:
            end = int(round((i["x1"] / ctc.sample_rate) * 75))
            word_tokens = full_codes[0][0][start:end]
            start = end
            if not word_tokens:
                word_tokens = [1]

            duration = round(len(word_tokens) / 75, 2)

            data.append({
                "word": i["word"],
                "duration": duration,
                "codes": word_tokens
            })

        return {
            "text": transcript,
            "words": data,
        }
    
    def generate(self, text: str, speaker: dict = None, temperature: float = 0.1, repetition_penalty: float = 1.1, max_lenght: int = 4096) -> ModelOutput:
        input_ids = self.prepare_prompt(text, speaker)
        output = self.model.generate(
            input_ids=input_ids,
            config=GenerationConfig(
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                max_length=max_lenght
            )
        )
        audio = self.get_audio(output[input_ids.size()[-1]:])
        return ModelOutput(audio, self.audio_codec.sr)

class InterfaceGGUF(InterfaceHF):
    def __init__(
            self,
            model_path: str,
            device: str = None,
            n_gpu_layers: int = 0,
            additional_model_config: dict = {}
    ) -> None:
        self.device = torch.device(
            device if device is not None 
            else "cuda" if torch.cuda.is_available() 
            else "cpu"
        )
        self.audio_codec = AudioCodec(self.device)
        self.prompt_processor = PromptProcessor("OuteAI/OuteTTS-0.1-350M")
        self.model = GGUFModel(
            model_path=model_path,
            n_gpu_layers=n_gpu_layers,
            additional_model_config=additional_model_config
        )

    def prepare_prompt(self, text: str, speaker: dict = None):
        prompt = self.prompt_processor.get_completion_prompt(text, speaker)
        return self.prompt_processor.tokenizer.encode(
            prompt, add_special_tokens=False)

    def generate(self, text: str, speaker: dict = None, temperature: float = 0.1, repetition_penalty: float = 1.1, max_lenght: int = 4096) -> ModelOutput:
        output = self.model.generate(
            input_ids=self.prepare_prompt(text, speaker),
            config=GenerationConfig(
                temperature=temperature,
                repetition_penalty=repetition_penalty,
                max_length=max_lenght
            )
        )
        audio = self.get_audio(output)
        return ModelOutput(audio, self.audio_codec.sr)
        



from dataclasses import dataclass
import torch
from loguru import logger
from transformers import AutoModelForCausalLM, AutoTokenizer

try:
    from llama_cpp import Llama, llama_token_is_eog
    _GGUF_AVAILABLE = True
except ImportError:
    _GGUF_AVAILABLE = False

@dataclass
class GenerationConfig:
    temperature: float = 0.1
    repetition_penalty: float = 1.1
    max_length: int = 4096

class HFModel:
    def __init__(
        self,
        model_path: str,
        device: str = None,
        dtype: torch.dtype = None,
        additional_model_config: dict = {}
    ) -> None:
        self.device = torch.device(
            device if device is not None 
            else "cuda" if torch.cuda.is_available() 
            else "cpu"
        )
        self.device = torch.device(device)
        self.dtype = dtype
        self.model = AutoModelForCausalLM.from_pretrained(
            model_path,
            torch_dtype=dtype,
            **additional_model_config
        ).to(device)

    def generate(self, input_ids: torch.Tensor, config: GenerationConfig) -> list[int]:
        return self.model.generate(
            input_ids,
            max_length=config.max_length,
            temperature=config.temperature,
            repetition_penalty=config.repetition_penalty,
            do_sample=True
        )[0].tolist()

class GGUFModel:
    def __init__(
            self,
            model_path: str,
            n_gpu_layers: int = 0,
            additional_model_config: dict = {}
    ) -> None:
        
        if not _GGUF_AVAILABLE:
            raise ImportError(
                "llama_cpp python module not found."
                "To use the GGUF model you must install llama cpp python manually."
            )

        additional_model_config["n_ctx"] = 4096
        self.model = Llama(
            model_path=model_path,
            n_gpu_layers=n_gpu_layers,
            **additional_model_config
        )
    
    def generate(self, input_ids: list[int], config: GenerationConfig) -> list[int]:
        tokens = []
        for token in self.model.generate(
            input_ids,
            temp=config.temperature,
            repeat_penalty=config.repetition_penalty
        ):
            tokens.append(token)
            if (llama_token_is_eog(self.model._model.model, token) or 
                len(tokens) >= config.max_length):
                break

        return tokens


import torchaudio
import torch
from .encoder.utils import convert_audio
from .decoder.pretrained import WavTokenizer
import requests
from tqdm import tqdm
import os
import platform
from loguru import logger

class AudioCodec:
    # WavTokenizer implementation: https://github.com/jishengpeng/WavTokenizer

    def __init__(self, device: str = None):
        self.device = torch.device(device if device is not None else "cuda" if torch.cuda.is_available() else "cpu")
        self.cache_dir = self.get_cache_dir()
        self.model_path = os.path.join(self.cache_dir, "wavtokenizer_large_speech_320_24k.ckpt")
        self.config_path = self.get_config_path()
        self.model_url = "https://huggingface.co/novateur/WavTokenizer-large-speech-75token/resolve/main/wavtokenizer_large_speech_320_24k.ckpt"
        self.ensure_model_exists()
        self.wavtokenizer = WavTokenizer.from_pretrained0802(self.config_path, self.model_path)
        self.wavtokenizer = self.wavtokenizer.to(self.device)
        self.sr = 24000
        self.bandwidth_id = torch.tensor([0])

    def get_config_path(self):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(current_dir, "wavtokenizer_config.yaml")

    def get_cache_dir(self):
        return os.path.join(
            os.getenv('APPDATA') if platform.system() == "Windows" else os.path.join(os.path.expanduser("~"), ".cache"), 
            "outeai", "tts", "wavtokenizer_large_speech_75_token")

    def ensure_model_exists(self):
        if not os.path.exists(self.model_path):
            logger.info(f"Downloading WavTokenizer model from {self.model_url}")
            os.makedirs(os.path.dirname(self.model_path), exist_ok=True)
            self.download_file(self.model_path, self.model_url)

    def download_file(self, save_path: str, url: str):
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get('content-length', 0))
        with open(save_path, 'wb') as file, tqdm(
            desc=save_path,
            total=total_size,
            unit='iB',
            unit_scale=True,
            unit_divisor=1024,
        ) as bar:
            for data in response.iter_content(chunk_size=1024):
                size = file.write(data)
                bar.update(size)

    def load_audio(self, path):
        wav, sr = torchaudio.load(path)
        wav = convert_audio(wav, sr, self.sr, 1) 
        return wav.to(self.device)

    def encode(self, audio: torch.Tensor):
        _,discrete_code= self.wavtokenizer.encode_infer(audio, bandwidth_id=torch.tensor([0]).to(self.device))
        return discrete_code
        
    def decode(self, codes):
        features = self.wavtokenizer.codes_to_features(codes)
        audio_out = self.wavtokenizer.decode(features, bandwidth_id=torch.tensor([0]).to(self.device))
        return audio_out

    def save_audio(self, audio: torch.Tensor, path: str):
        torchaudio.save(path, audio.cpu(), sample_rate=self.sr, encoding='PCM_S', bits_per_sample=16)